
import os
import numpy as np
import h5py


def count_number_of_events_in_folder(dirpath):

    fpath_list = []
    for file in os.listdir(dirpath):
        if file.endswith('.h5'):
            fpath = os.path.join(dirpath, file)
            fpath_list.append(fpath)

    n_total = 0
    for fpath in fpath_list:
        f = h5py.File(fpath, 'r')
        n_total += f['event_info'].shape[0]
        f.close()

    return n_total



def main():
    dirpath_rn = '/home/saturn/capn/mppi033h/Data/raw_data/random_noise'
    rn_n_total = count_number_of_events_in_folder(dirpath_rn)
    print('--------------------------------------------------------')
    print('Total number of random_noise events: ' + str(rn_n_total))
    print('--------------------------------------------------------')

    dirpath_mupage = '/home/saturn/capn/mppi033h/Data/raw_data/mupage'
    mupage_n_total = count_number_of_events_in_folder(dirpath_mupage)
    print('--------------------------------------------------------')
    print('Total number of mupage events: ' + str(mupage_n_total))
    print('--------------------------------------------------------')

    dirpath_muon_cc_low_e = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/1-5GeV/muon-CC'
    dirpath_muon_cc_high_e = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/muon-CC'
    muon_cc_low_e_n_total = count_number_of_events_in_folder(dirpath_muon_cc_low_e)
    muon_cc_high_e_n_total = count_number_of_events_in_folder(dirpath_muon_cc_high_e)
    print('--------------------------------------------------------')
    print('Total number of muon-CC 1-5GeV events: ' + str(muon_cc_low_e_n_total))
    print('Total number of muon-CC 3-100GeV events: ' + str(muon_cc_high_e_n_total))
    print('--------------------------------------------------------')

    dirpath_elec_cc_low_e = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/1-5GeV/elec-CC'
    dirpath_elec_cc_high_e = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/elec-CC'
    elec_cc_low_e_n_total = count_number_of_events_in_folder(dirpath_elec_cc_low_e)
    elec_cc_high_e_n_total = count_number_of_events_in_folder(dirpath_elec_cc_high_e)
    print('--------------------------------------------------------')
    print('Total number of elec-CC 1-5GeV events: ' + str(elec_cc_low_e_n_total))
    print('Total number of elec-CC 3-100GeV events: ' + str(elec_cc_high_e_n_total))
    print('--------------------------------------------------------')

    dirpath_elec_nc_low_e = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/1-5GeV/elec-NC'
    dirpath_elec_nc_high_e = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/elec-NC'
    elec_nc_low_e_n_total = count_number_of_events_in_folder(dirpath_elec_nc_low_e)
    elec_nc_high_e_n_total = count_number_of_events_in_folder(dirpath_elec_nc_high_e)
    print('--------------------------------------------------------')
    print('Total number of elec-NC 1-5GeV events: ' + str(elec_nc_low_e_n_total))
    print('Total number of elec-NC 3-100GeV events: ' + str(elec_nc_high_e_n_total))
    print('--------------------------------------------------------')

    dirpath_tau_cc = '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/tau-CC'
    tau_cc_n_total = count_number_of_events_in_folder(dirpath_tau_cc)
    print('--------------------------------------------------------')
    print('Total number of tau-CC 3-100GeV events: ' + str(tau_cc_n_total))
    print('--------------------------------------------------------')


    n_total_neutr = muon_cc_low_e_n_total + muon_cc_high_e_n_total + elec_cc_low_e_n_total + elec_cc_high_e_n_total\
                    + elec_nc_low_e_n_total + elec_nc_high_e_n_total + tau_cc_n_total
    print('--------------------------------------------------------')
    print('Total number of neutrino events: ' + str(n_total_neutr))
    print('--------------------------------------------------------')


if __name__ == '__main__':
    main()



import os
import h5py
import km3pipe as kp
import km3modules as km

import orcasong
import orcasong.modules as modules
import orcasong.plotting.plot_binstats as plot_binstats
from orcasong.mc_info_extr import get_mc_info_extr


__author__ = 'Stefan Reck'


class FileBinner:
    """
    For making binned images and mc_infos, which can be used for conv. nets.

    Can also add statistics of the binning to the h5 files, which can
    be plotted to show the distribution of hits among the bins and how
    many hits were cut off.

    Attributes
    ----------
    n_statusbar : int, optional
        Print a statusbar every n blobs.
    n_memory_observer : int, optional
        Print memory usage every n blobs.
    complib : str
        Compression library used for saving the output to a .h5 file.
        All PyTables compression filters are available, e.g. 'zlib',
        'lzf', 'blosc', ... .
    complevel : int
        Compression level for the compression filter that is used for
        saving the output to a .h5 file.
    flush_frequency : int
        After how many events the accumulated output should be flushed to
        the harddisk.
        A larger value leads to a faster orcasong execution,
        but it increases the RAM usage as well.
    bin_plot_freq : int or None
        If int is given, defines after how many blobs data for an overview
        histogram is extracted.
        It shows the distribution of hits, the bin edges, and how many hits
        were cut off for each field name in bin_edges_list.
        It will be saved to the same path as the outfile in run.

    """
    def __init__(self,
                 bin_edges_list,
                 mc_info_extr=None,
                 det_file=None,
                 center_time=True,
                 event_skipper=None,
                 add_bin_stats=True,
                 chunksize=32,
                 keep_event_info=True,
                 keep_mc_tracks=False,
                 add_t0=False,):
        """
        Parameters
        ----------
        bin_edges_list : List
            List with the names of the fields to bin, and the respective bin
            edges, including the left- and right-most bin edge.
            Example: For 10 bins in the z direction, and 100 bins in time:
                bin_edges_list = [
                    ["pos_z", np.linspace(0, 10, 11)],
                    ["time", np.linspace(-50, 550, 101)],
                ]
            Some examples can be found in orcasong.bin_edges.
        mc_info_extr : function, optional
            Function that extracts desired mc_info from a blob, which is then
            stored as the "y" datafield in the .h5 file.
            The function takes the km3pipe blob as an input, and returns
            a dict mapping str to floats.
            Some examples can be found in orcasong.mc_info_extr.
        det_file : str, optional
            Path to a .detx detector geometry file, which can be used to
            calibrate the hits.
        center_time : bool
            Subtract time of first triggered hit from all hit times. Will
            also be done for McHits if they are in the blob [default: True].
        event_skipper : func, optional
            Function that takes the blob as an input, and returns a bool.
            If the bool is true, the blob will be skipped.
        add_bin_stats : bool
            Add statistics of the binning to the output file. They can be
            plotted with util/bin_stats_plot.py [default: True].
        chunksize : int
            Chunksize (along axis_0) used for saving the output
            to a .h5 file [default: 32].
        keep_event_info : bool
            If True, will keep the "event_info" table [default: True].
        keep_mc_tracks : bool
            If True, will keep the "McTracks" table [default: False].
        add_t0 : bool
            If true, add t0 to the time of hits. If using a det_file,
            this will already have been done automatically [default: False].

        """
        self.bin_edges_list = bin_edges_list
        self.mc_info_extr = mc_info_extr
        self.det_file = det_file
        self.add_t0 = add_t0
        self.center_time = center_time
        self.event_skipper = event_skipper

        self.keep_event_info = keep_event_info
        self.keep_mc_tracks = keep_mc_tracks
        self.chunksize = chunksize

        if add_bin_stats:
            self.bin_plot_freq = 1
        else:
            self.bin_plot_freq = None

        self.n_statusbar = 1000
        self.n_memory_observer = 1000
        self.complib = 'zlib'
        self.complevel = 1
        self.flush_frequency = 1000

    def run(self, infile, outfile=None, save_plot=False):
        """
        Generate images from the infile, and save them as the outfile.

        Parameters
        ----------
        infile : str
            Path to the input file.
        outfile : str, optional
            Path to the output file (will be created). If none is given,
            will auto generate the name and save it in the cwd.
        save_plot : bool
            Save the binning hists as a pdf. Only possible if add_bin_stats
            is True.

        """
        if save_plot and self.bin_plot_freq is None:
            raise ValueError("Can not make plot when add_bin_stats is False")

        name, shape = self.get_names_and_shape()
        print("Generating {} images with shape {}".format(name, shape))

        if outfile is None:
            infile_basename = os.path.basename(infile)
            outfile_name = os.path.splitext(infile_basename)[0] + "_binned.h5"
            outfile = os.path.join(os.getcwd(), outfile_name)

        pipe = self.build_pipe(infile, outfile)
        smry = pipe.drain()

        if self.bin_plot_freq is not None:
            hists = smry["BinningStatsMaker"]
            plot_binstats.add_hists_to_h5file(hists, outfile)

            if save_plot:
                save_as = os.path.splitext(outfile)[0] + "_hists.pdf"
                plot_binstats.plot_hists(hists, save_as)

        add_version_info(outfile)

    def run_multi(self, infiles, outfolder, save_plot=False):
        """
        Bin multiple files into their own output files each.
        The output file names will be generated automatically.

        Parameters
        ----------
        infiles : List
            The path to infiles as str.
        outfolder : str
            The output folder to place them in.
        save_plot : bool
            Save the binning hists as a pdf. Only possible if add_bin_stats
            is True.

        """
        if save_plot and self.bin_plot_freq is None:
            raise ValueError("Can not make plot when add_bin_stats is False")

        outfiles = []
        for infile in infiles:
            outfile_name = os.path.splitext(os.path.basename(infile))[0] \
                           + "_hist.h5"
            outfile = os.path.join(outfolder, outfile_name)
            outfiles.append(outfile)

            self.run(infile, outfile, save_plot=False)

        if save_plot:
            plot_binstats.plot_hist_of_files(
                files=outfiles, save_as=outfolder+"binning_hist.pdf")

    def build_pipe(self, infile, outfile):
        """
        Build the pipeline to generate images and mc_info for a file.
        """
        pipe = kp.Pipeline()

        if self.n_statusbar is not None:
            pipe.attach(km.common.StatusBar, every=self.n_statusbar)
        if self.n_memory_observer is not None:
            pipe.attach(km.common.MemoryObserver, every=self.n_memory_observer)

        pipe.attach(kp.io.hdf5.HDF5Pump, filename=infile)
        pipe.attach(km.common.Keep, keys=['EventInfo', 'Header', 'RawHeader',
                                          'McTracks', 'Hits', 'McHits'])

        if self.det_file:
            pipe.attach(modules.DetApplier, det_file=self.det_file)

        if self.center_time or self.add_t0:
            pipe.attach(modules.TimePreproc,
                        add_t0=self.add_t0,
                        center_time=self.center_time)

        if self.event_skipper is not None:
            pipe.attach(modules.EventSkipper, event_skipper=self.event_skipper)

        if self.bin_plot_freq is not None:
            pipe.attach(modules.BinningStatsMaker,
                        bin_plot_freq=self.bin_plot_freq,
                        bin_edges_list=self.bin_edges_list)

        pipe.attach(modules.ImageMaker,
                    bin_edges_list=self.bin_edges_list,
                    store_as="histogram")

        if self.mc_info_extr is not None:
            if isinstance(self.mc_info_extr, str):
                mc_info_extr = get_mc_info_extr(self.mc_info_extr)
            else:
                mc_info_extr = self.mc_info_extr

            pipe.attach(modules.McInfoMaker,
                        mc_info_extr=mc_info_extr,
                        store_as="mc_info")

        keys_keep = ['histogram', 'mc_info']
        if self.keep_event_info:
            keys_keep.append('EventInfo')
        if self.keep_mc_tracks:
            keys_keep.append('McTracks')
        pipe.attach(km.common.Keep, keys=keys_keep)

        pipe.attach(kp.io.HDF5Sink,
                    filename=outfile,
                    complib=self.complib,
                    complevel=self.complevel,
                    chunksize=self.chunksize,
                    flush_frequency=self.flush_frequency)
        return pipe

    def get_names_and_shape(self):
        """
        Get names and shape of the resulting x data,
        e.g. (pos_z, time), (18, 50).
        """
        names, shape = [], []
        for bin_name, bin_edges in self.bin_edges_list:
            names.append(bin_name)
            shape.append(len(bin_edges) - 1)

        return tuple(names), tuple(shape)

    def __repr__(self):
        name, shape = self.get_names_and_shape()
        return "<FileBinner: {} {}>".format(name, shape)


def add_version_info(file):
    """ Add current orcasong version to h5 file. """
    with h5py.File(file, "a") as f:
        f.attrs.create("orcasong", orcasong.__version__, dtype="S6")



"""
Functions that extract info from a blob for the mc_info / y datafield
in the h5 files.

These are made for the specific given runs. They might not be
applicable to other data, and could cause errors or produce unexpected
results when used on data other then the specified.

"""

import warnings
import numpy as np

__author__ = 'Stefan Reck'


def get_mc_info_extr(mc_info_extr):
    """
    Get an existing mc info extractor function.

    Attributes
    ----------
    mc_info_extr : function
        Function to extract the info. Takes the blob as input, outputs
        a dict with the desired mc_infos.

    """
    if mc_info_extr == "mupage":
        funct = "get_mupage_mc"
        mc_info_extr = get_mupage_mc

    elif mc_info_extr == "real_data":
        funct = "get_real_data"
        mc_info_extr = get_real_data

    elif mc_info_extr == "random_noise":
        funct = "get_pure_noise"
        mc_info_extr = get_pure_noise

    else:
        raise NameError("Unknown mc_info_type " + mc_info_extr)

    # TODO deprecated
    wrng = "The use of a str for mc_info_extr is deprecated. Import the " \
           "function {} from orcasong.mc_info_types instead, and use this" \
           "as mc_info_extr".format(funct)
    warnings.warn(wrng)

    return mc_info_extr


def get_real_data(blob):
    """
    Get info present in real data.
    Designed for the 2017 one line runs.

    """
    event_info = blob['EventInfo']

    track = {
        'event_id': event_info.event_id,
        # was .event_id[0] up to km3pipe 8.16.0
        'run_id': event_info.run_id,
        'trigger_mask': event_info.trigger_mask,
    }
    return track


def get_pure_noise(blob):
    """
    For simulated pure noise events, which have particle_type 0.

    """
    event_info = blob['EventInfo']

    track = {
        'event_id': event_info.event_id[0],
        'run_id': event_info.run_id,
        'particle_type': 0
    }
    return track


def get_mupage_mc(blob):
    """
    For mupage muon simulations.

    Will only take into account muons with at least 1 McHit in the active
    line of the detector.

    Designed for the 2017 run by run mupage simulations.
    e.g. mcv5.1_r3.mupage_10G.km3_AAv1.jterbr00002800.5103.root.h5

    Parameters
    ----------
    blob : dict
        The blob from the pipeline.

    Returns
    -------
    track : dict
        The info for mc_info.

    """
    # only one line has hits, but there are two for the mc. This one is active:
    active_du = 2

    track = dict()

    # total number of simulated muons, not all might deposit counts
    n_muons_sim = blob['McTracks'].shape[0]
    track["n_muons_sim"] = n_muons_sim

    # Origin of each mchit (as int) in the active line
    in_active_du = blob["McHits"]["du"] == active_du
    origin = blob["McHits"]["origin"][in_active_du]
    # get how many mchits were produced per muon in the bundle
    origin_dict = dict(zip(*np.unique(origin, return_counts=True)))
    origin_list = []
    for i in range(1, n_muons_sim + 1):
        origin_list.append(origin_dict.get(i, 0))
    # origin_list[i] is num of mc_hits of muon i in active du
    origin_list = np.array(origin_list)

    visible_mc_tracks = blob["McTracks"][origin_list > 0]
    visible_origin_list = origin_list[origin_list > 0]

    # only muons with at least one mchit in active line
    n_muons = len(visible_origin_list)
    if n_muons == 0:
        warnings.warn("No visible muons in blob!")

    track["n_muons"] = n_muons

    track["event_id"] = blob['EventInfo'].event_id[0]
    track["run_id"] = blob["EventInfo"].run_id[0]
    # run_id = blob['Header'].start_run.run_id.astype('float32')

    # take 0: assumed that this is the same for all muons in a bundle
    track["particle_type"] = visible_mc_tracks[0].type if n_muons != 0 else 0

    # always 1 actually
    # track["is_cc"] = blob['McTracks'][0].is_cc

    # always 0 actually
    # track["bjorkeny"] = blob['McTracks'][0].bjorkeny

    # same for all muons in a bundle TODO not?
    track["time_interaction"] = visible_mc_tracks[0].time if n_muons != 0 else 0

    # sum up the energy of all visible muons
    energy = visible_mc_tracks.energy if n_muons != 0 else 0
    track["energy"] = np.sum(energy)

    track["n_mc_hits"] = np.sum(visible_origin_list)
    desc_order = np.argsort(-visible_origin_list)

    # Sort by energy, highest first
    sorted_energy = energy[desc_order] if n_muons != 0 else []
    sorted_mc_hits = visible_origin_list[desc_order] if n_muons != 0 else []

    # Store number of mchits of the 10 highest mchits muons (-1 if it has less)
    for i in range(10):
        field_name = "n_mc_hits_"+str(i)

        if i < len(sorted_mc_hits):
            field_data = sorted_mc_hits[i]
        else:
            field_data = -1

        track[field_name] = field_data

    # Store energy of the 10 highest mchits muons (-1 if it has less)
    for i in range(10):
        field_name = "energy_"+str(i)

        if i < len(sorted_energy):
            field_data = sorted_energy[i]
        else:
            field_data = -1

        track[field_name] = field_data

    # only muons with at least 5 mchits in active line
    track["n_muons_5_mchits"] = len(origin_list[origin_list > 4])
    # only muons with at least 10 mchits in active line
    track["n_muons_10_mchits"] = len(origin_list[origin_list > 9])

    # all muons in a bundle are parallel, so just take dir of first muon
    track["dir_x"] = visible_mc_tracks[0].dir_x if n_muons != 0 else 0
    track["dir_y"] = visible_mc_tracks[0].dir_y if n_muons != 0 else 0
    track["dir_z"] = visible_mc_tracks[0].dir_z if n_muons != 0 else 0

    if n_muons != 0:
        # vertex is the weighted (energy) mean of the individual vertices
        track["vertex_pos_x"] = np.average(visible_mc_tracks[:].pos_x,
                                           weights=visible_mc_tracks[:].energy)
        track["vertex_pos_y"] = np.average(visible_mc_tracks[:].pos_y,
                                           weights=visible_mc_tracks[:].energy)
        track["vertex_pos_z"] = np.average(visible_mc_tracks[:].pos_z,
                                           weights=visible_mc_tracks[:].energy)
    else:
        track["vertex_pos_x"] = 0
        track["vertex_pos_y"] = 0
        track["vertex_pos_z"] = 0

    return track



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
For investigating the ideal binning, based on the info in calibrated
.h5 files.

Specialized classes TimePlotter and ZPlotter are available for plotting
the time/ Z-Coordinate.

"""

import numpy as np
import km3pipe as kp
import matplotlib.pyplot as plt

from orcasong.modules import TimePreproc, DetApplier

__author__ = 'Stefan Reck'


class FieldPlotter:
    """
    Baseclass for investigating the ideal binning, based on the info in
    a field of calibrated .h5 files.

    Intended for 1d binning, like for fields "time" or "pos_z".
    Workflow:
    1. Initialize with files, then run .plot() to extract and store
       the data, and show the plot interactively.
    2. Choose a binning via .set_binning.
    3. Run .plot() again to show the plot with the adjusted binning on the
       stored data.
    4. Repeat step 2 and 3 unitl happy with binning.
    (5.) Save plot via .plot(savepath), or get the bin edges via .get_bin_edges()

    The plot will have some bins attached in both directions for
    better overview.

    Attributes:
    -----------
    files : list or str
        The .h5 file(s).
    field : str
        The field to look stuff up, e.g. "time", "pos_z", ...
    det_file : str, optional
        Path to a detx file for calibrating.
    filter_for_du : int, optional
        Only get hits from one specific du, specified by the int.
    hits : ndarray
        The extracted Hits.
    mc_hits : ndarray
        The extracted McHits, if present.
    n_events : int
        The number of events in the extracted data.
    limits : List
        Left- and right-most edge of the binning.
    n_bins : int
        The number of bins.
    plot_padding : List
        Fraction of bins to append to left and right direction
        (only in the plot for better overview).
    x_label : str
        X label of the plot.
    y_label : str
        Y label of the plot.
    hist_kwargs : dict
        Kwargs for plt.hist
    xlim : List
        The xlimits of the hist plot.
    show_plots : bool
        If True, auto plt.show() the plot.

    """
    def __init__(self, files, field, det_file=None):
        self.files = files
        self.field = field
        self.filter_for_du = None
        self.det_file = det_file

        self.hits = None
        self.mc_hits = None
        self.n_events = None

        self.limits = None
        self.n_bins = 100
        self.plot_padding = [0.2, 0.2]

        # Plotting stuff
        self.xlabel = None
        self.ylabel = 'Fraction of hits'
        self.hist_kwargs = {
            "histtype": "stepfilled",
            "density": True,
        }
        self.xlim = None
        self.ylim = None
        self.show_plots = True
        self.last_ylim = None

    def plot(self, only_mc_hits=False, save_path=None):
        """
        Generate and store or load the data, then make the plot.

        Parameters
        ----------
        only_mc_hits : bool
            If true, plot the McHits instead of the Hits.
        save_path : str, optional
            Save plot to here.

        Returns
        -------
        fig, ax : pyplot figure
            The plot.

        """
        if self.hits is None:
            self.extract()
        fig, ax = self.make_histogram(only_mc_hits, save_path)
        return fig, ax

    def set_binning(self, limits, n_bins):
        """
        Set the desired binning.

        Parameters
        ----------
        limits : List
            Left- and right-most edge of the binning.
        n_bins : int
            The number of bins.

        """
        self.limits = limits
        self.n_bins = n_bins

    def get_binning(self):
        """
        Set the desired binning.

        Returns
        -------
        limits : List
            Left- and right-most edge of the binning.
        n_bins : int
            The number of bins.

        """
        return self.limits, self.n_bins

    def get_bin_edges(self):
        """
        Get the bin edges as a ndarray.

        """
        limits, n_bins = self.get_binning()

        if limits is None:
            raise ValueError("Can not return bin edges: No binning limits set")

        bin_edges = np.linspace(limits[0], limits[1], n_bins + 1)

        return bin_edges

    def extract(self):
        """
        Extract the content of a field from all events in the file(s) and
        store it.

        """
        data_all_events = None
        mc_all_events = None
        self.n_events = 0

        if not isinstance(self.files, list):
            files = [self.files]
        else:
            files = self.files

        event_pump = kp.io.hdf5.HDF5Pump(filenames=files)

        if self.det_file is not None:
            cal = DetApplier(det_file=self.det_file)

        for i, event_blob in enumerate(event_pump):
            self.n_events += 1

            if self.det_file is not None:
                event_blob = cal.process(event_blob)

            if i % 2000 == 0:
                print("Blob no. "+str(i))

            data_one_event = self._get_hits(event_blob, get_mc_hits=False)

            if data_all_events is None:
                data_all_events = data_one_event
            else:
                data_all_events = np.concatenate(
                    [data_all_events, data_one_event], axis=0)

            if "McHits" in event_blob:
                mc_one_event = self._get_hits(event_blob, get_mc_hits=True)

                if mc_all_events is None:
                    mc_all_events = mc_one_event
                else:
                    mc_all_events = np.concatenate(
                        [mc_all_events, mc_one_event], axis=0)

        event_pump.close()

        print("Number of events: " + str(self.n_events))

        self.hits = data_all_events
        self.mc_hits = mc_all_events

    def make_histogram(self, only_mc_hits=False, save_path=None):
        """
        Plot the hist data. Can also save it if given a save path.

        Parameters
        ----------
        only_mc_hits : bool
            If true, plot the McHits instead of the Hits.
        save_path : str, optional
            Save the fig to this path.

        Returns
        -------
        fig, ax : pyplot figure
            The plot.

        """
        if only_mc_hits:
            data = self.mc_hits
        else:
            data = self.hits

        if data is None:
            raise ValueError("Can not make histogram, no data extracted yet.")

        bin_edges = self._get_padded_bin_edges()

        fig, ax = plt.subplots()
        bins = plt.hist(data, bins=bin_edges, **self.hist_kwargs)[1]
        print("Size of first bin: " + str(bins[1] - bins[0]))

        plt.grid(True, zorder=0, linestyle='dotted')

        if self.limits is not None:
            for bin_line_x in self.limits:
                plt.axvline(x=bin_line_x, color='firebrick', linestyle='--')

        if self.xlabel is None:
            plt.xlabel(self._get_xlabel())

        if self.xlim is not None:
            plt.xlim(self.xlim)

        if self.ylim is not None:
            plt.ylim(self.ylim)

        plt.ylabel(self.ylabel)
        plt.tight_layout()

        if save_path is not None:
            print("Saving plot to "+str(save_path))
            plt.savefig(save_path)

        if self.show_plots:
            plt.show()

        return fig, ax

    def _get_padded_bin_edges(self):
        """
        Get the padded bin edges.

        """
        limits, n_bins = self.get_binning()

        if limits is None:
            bin_edges = n_bins

        else:
            total_range = limits[1] - limits[0]
            bin_size = total_range / n_bins

            addtnl_bins = [
                int(self.plot_padding[0] * n_bins),
                int(self.plot_padding[1] * n_bins)
            ]

            padded_range = [
                limits[0] - bin_size * addtnl_bins[0],
                limits[1] + bin_size * addtnl_bins[1]
            ]
            padded_n_bins = n_bins + addtnl_bins[0] + addtnl_bins[1]
            bin_edges = np.linspace(padded_range[0], padded_range[1],
                                    padded_n_bins + 1)

        return bin_edges

    def _get_hits(self, blob, get_mc_hits):
        """
        Get desired attribute from an event blob.

        Parameters
        ----------
        blob
            The blob.
        get_mc_hits : bool
            If true, will get the "McHits" instead of the "Hits".

        Returns
        -------
        blob_data : ndarray
            The data.

        """
        if get_mc_hits:
            field_name = "McHits"
        else:
            field_name = "Hits"

        blob_data = blob[field_name][self.field]

        if self.filter_for_du is not None:
            dus = blob[field_name]["du"]
            blob_data = blob_data[dus == self.filter_for_du]

        return blob_data

    def _get_xlabel(self):
        """
        Some saved xlabels.

        """
        if self.field == "time":
            xlabel = "Time [ns]"
        elif self.field == "pos_z":
            xlabel = "Z position [m]"
        else:
            xlabel = None
        return xlabel

    def __repr__(self):
        return "<FieldPlotter: {}>".format(self.files)


class TimePlotter(FieldPlotter):
    """
    For plotting the time.
    """
    def __init__(self, files, add_t0=True, det_file=None):
        field = "time"
        FieldPlotter.__init__(self, files, field, det_file=det_file)
        self.tp = TimePreproc(add_t0=add_t0)

    def _get_hits(self, blob, get_mc_hits):
        blob = self.tp.process(blob)

        if get_mc_hits:
            field_name = "McHits"
        else:
            field_name = "Hits"

        blob_data = blob[field_name][self.field]

        if self.filter_for_du is not None:
            dus = blob[field_name]["du"]
            blob_data = blob_data[dus == self.filter_for_du]

        return blob_data


class ZPlotter(FieldPlotter):
    """
    For plotting the z dim.
    """
    def __init__(self, files, det_file=None):
        field = "pos_z"
        FieldPlotter.__init__(self, files, field, det_file=det_file)

        self.plotting_bins = 100

    def _get_padded_bin_edges(self):
        """
        Get the padded bin edges.

        """
        return self.plotting_bins

    def set_binning(self, limits, n_bins):
        """
        Set the desired binning.

        Parameters
        ----------
        limits : List
            Left- and right-most edge of the binning.
        n_bins : int
            The number of bins.

        """
        bin_edges = np.linspace(limits[0], limits[1],
                                n_bins + 1)
        self.limits = bin_edges
        self.n_bins = n_bins

    def get_bin_edges(self):
        return self.limits



"""
Custom km3pipe modules for making nn input files.
"""

import warnings
import numpy as np
import km3pipe as kp

__author__ = 'Stefan Reck'


class McInfoMaker(kp.Module):
    """
    Get the desired mc_info from the blob.

    Attributes
    ----------
    mc_info_extr : function
        Function to extract the info. Takes the blob as input, outputs
        a dict with the desired mc_infos.
    store_as : str
        Store the mcinfo with this name in the blob.

    """

    def configure(self):
        self.mc_info_extr = self.require('mc_info_extr')
        self.store_as = self.require('store_as')

    def process(self, blob):
        track = self.mc_info_extr(blob)
        dtypes = [(key, np.float64) for key in track.keys()]
        kp_hist = kp.dataclasses.Table(
            track, dtype=dtypes,  h5loc='y', name='event_info')

        blob[self.store_as] = kp_hist
        return blob


class TimePreproc(kp.Module):
    """
    Preprocess the time in the blob.

    Can add t0 to hit times.
    Times of hits and mchits can be centered with the time of the first
    triggered hit.

    Attributes
    ----------
    add_t0 : bool
        If true, t0 will be added.
    center_time : bool
        If true, center hit and mchit times.

    """

    def configure(self):
        self.add_t0 = self.require('add_t0')
        self.center_time = self.get('center_time', default=True)

        self.has_mchits = None
        self._t0_flag = False
        self._cent_hits_flag = False
        self._cent_mchits_flag = False

    def process(self, blob):
        if self.has_mchits is None:
            self.has_mchits = "McHits" in blob

        if self.add_t0:
            blob = self.add_t0_time(blob)
        blob = self.center_hittime(blob)

        return blob

    def add_t0_time(self, blob):
        if not self._t0_flag:
            self._t0_flag = True
            print("Adding t0 to hit times")
        blob["Hits"].time = np.add(blob["Hits"].time, blob["Hits"].t0)

        if self.has_mchits:
            blob["McHits"].time = np.add(blob["McHits"].time,
                                         blob["McHits"].t0)

        return blob

    def center_hittime(self, blob):
        hits_time = blob["Hits"].time
        hits_triggered = blob["Hits"].triggered
        t_first_trigger = np.min(hits_time[hits_triggered == 1])

        if self.center_time:
            if not self._cent_hits_flag:
                print("Centering time of Hits")
                self._cent_hits_flag = True
            blob["Hits"].time = np.subtract(hits_time, t_first_trigger)

        if self.has_mchits:
            if not self._cent_mchits_flag:
                print("Centering time of McHits")
                self._cent_mchits_flag = True
            mchits_time = blob["McHits"].time
            blob["McHits"].time = np.subtract(mchits_time, t_first_trigger)

        return blob


class ImageMaker(kp.Module):
    """
    Make a n-d histogram from "Hits" in blob, and store it.

    Attributes
    ----------
    bin_edges_list : List
        List with the names of the fields to bin, and the respective bin edges,
        including the left- and right-most bin edge.
    store_as : str
        Store the images with this name in the blob.

    """

    def configure(self):
        self.bin_edges_list = self.require('bin_edges_list')
        self.store_as = self.require('store_as')

    def process(self, blob):
        data, bins, name = [], [], ""

        for bin_name, bin_edges in self.bin_edges_list:
            data.append(blob["Hits"][bin_name])
            bins.append(bin_edges)
            name += bin_name + "_"

        histogram = np.histogramdd(data, bins=bins)[0]
        title = name + "event_images"

        hist_one_event = histogram[np.newaxis, ...].astype(np.uint8)
        kp_hist = kp.dataclasses.NDArray(
            hist_one_event, h5loc='x', title=title)

        blob[self.store_as] = kp_hist
        return blob


class BinningStatsMaker(kp.Module):
    """
    Generate a histogram of the number of hits for each binning field name.

    E.g. if the bin_edges_list contains "pos_z", this will make a histogram
    of #Hits vs. "pos_z", together with how many hits were outside
    of the bin edges in both directions.

    Per default, the resolution of the histogram (width of bins) will be
    higher then the given bin edges, and the edges will be stored seperatly.
    The time is the exception: The plotted bins have exactly the
    given bin edges.

    Attributes
    ----------
    bin_edges_list : List
        List with the names of the fields to bin, and the respective bin edges,
        including the left- and right-most bin edge.
    bin_plot_freq : int
        Extract data for the histograms only every given number of blobs
        (reduces time the pipeline takes to complete).
    res_increase : int
        Increase the number of bins by this much in the hists (so that one
        can see if the edges have been placed correctly). Is never used
        for the time binning (field name "time").

    """

    def configure(self):
        self.bin_edges_list = self.require('bin_edges_list')

        self.pdf_path = self.get('pdf_path', default=None)
        self.bin_plot_freq = self.get("bin_plot_freq", default=1)
        self.res_increase = self.get('res_increase', default=5)

        self.hists = {}
        for bin_name, org_bin_edges in self.bin_edges_list:
            # dont space bin edges for time
            if bin_name == "time":
                bin_edges = org_bin_edges
            else:
                bin_edges = self._space_bin_edges(org_bin_edges)

            self.hists[bin_name] = {
                "hist": np.zeros(len(bin_edges) - 1),
                "hist_bin_edges": bin_edges,
                "bin_edges": org_bin_edges,
                # below smallest edge, above largest edge:
                "cut_off": np.zeros(2),
            }

        self.i = 0

    def _space_bin_edges(self, bin_edges):
        """
        Increase resolution of given binning.
        """
        increased_n_bins = (len(bin_edges) - 1) * self.res_increase + 1
        bin_edges = np.linspace(
            bin_edges[0], bin_edges[-1], increased_n_bins)

        return bin_edges

    def process(self, blob):
        """
        Extract data from blob for the hist plots.
        """
        if self.i % self.bin_plot_freq == 0:
            for bin_name, hists_data in self.hists.items():
                hist_bin_edges = hists_data["hist_bin_edges"]

                hits = blob["Hits"]
                data = hits[bin_name]
                # get how much is cut off due to these limits
                out_pos = data[data > np.max(hist_bin_edges)].size
                out_neg = data[data < np.min(hist_bin_edges)].size

                # get all hits which are not cut off by other bin edges
                data = hits[bin_name][self._is_in_limits(
                    hits, excluded=bin_name)]
                hist = np.histogram(data, bins=hist_bin_edges)[0]

                self.hists[bin_name]["hist"] += hist
                self.hists[bin_name]["cut_off"] += np.array([out_neg, out_pos])

        self.i += 1
        return blob

    def finish(self):
        """
        Get the hists, which are the stats of the binning.

        Its a dict with each binning field name containing the following
        ndarrays:

        bin_edges : The actual bin edges.
        cut_off : How many events were cut off in positive and negative
            direction due to this binning.
        hist_bin_edges : The bin edges for the plot in finer resolution then
            the actual bin edges.
        hist : The number of hist in each bin of the hist_bin_edges.

        """
        return self.hists

    def _is_in_limits(self, hits, excluded=None):
        """ Get which hits are in the limits defined by ALL bin edges
        (except for given one). """
        inside = None
        for dfield, edges in self.bin_edges_list:
            if dfield == excluded:
                continue
            is_in = np.logical_and(hits[dfield] >= min(edges),
                                   hits[dfield] <= max(edges))
            if inside is None:
                inside = is_in
            else:
                inside = np.logical_and(inside, is_in)
        return inside


class EventSkipper(kp.Module):
    """
    Skip events based on some user function.

    Attributes
    ----------
    event_skipper : func
        Function that takes the blob as an input, and returns a bool.
        If the bool is true, the blob will be skipped.

    """

    def configure(self):
        self.event_skipper = self.require('event_skipper')

    def process(self, blob):
        skip_event = self.event_skipper(blob)
        if skip_event:
            return
        else:
            return blob


class DetApplier(kp.Module):
    """
    Apply calibration to the Hits with a detx file.

    Attributes
    ----------
    det_file : str
        Path to a .detx detector geometry file.

    """

    def configure(self):
        self.det_file = self.require("det_file")

        self.calib = kp.calib.Calibration(filename=self.det_file)
        self._calib_checked = False

        # for debugging
        self._assert_t0_is_added = False

    def process(self, blob):
        if self._calib_checked is False:
            if "pos_x" in blob["Hits"]:
                warnings.warn("Warning: Using a det file, but pos_x in Hits "
                              " detected. Is the file already "
                              "calibrated? This might lead to errors with t0.")
            self._calib_checked = True

        # original_time = blob["Hits"].time

        blob = self.calib.process(blob, key="Hits", outkey="Hits")
        if "McHits" in blob:
            blob = self.calib.process(blob, key="McHits", outkey="McHits")

        """
        actual_time = blob["Hits"].time
        t0 = blob["Hits"].t0
        target_time = np.add(original_time, t0)
        if not np.array_equal(actual_time, target_time):
            print(actual_time)
            print(target_time)
            raise AssertionError("t0 not added!")
        else:
            print("t0 was added ok")
        """
        return blob



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code that calculates the fraction of track to shower events for given files.
"""

import os
import numpy as np
import matplotlib as mpl
mpl.use('Agg')
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import pyplot as plt
import h5py
import natsort as ns


def get_h5_filepaths(dirpath):
    """
    Returns the filepaths of all .h5 files that are located in a specific directory.

    Parameters
    ----------
    dirpath: str
        Path of the directory where the .h5 files are located.

    Returns
    -------
    filepaths : list
        List with the full filepaths of all .h5 files in the dirpath folder.

    """
    filepaths = []
    for f in os.listdir(dirpath):
        if f.endswith('.h5'):
            filepaths.append(dirpath + '/' + f)

    filepaths = ns.natsorted(filepaths)  # TODO should not be necessary actually!
    return filepaths


def get_energies_for_fpaths(fpath_list, fpath_list_key_ic, cut_e_higher_than_3=False):
    """

    Parameters
    ----------
    fpath_list
    fpath_list_key_ic
    cut_e_higher_than_3

    Returns
    -------

    """

    energy_conc_arr = None
    for i, fpath in enumerate(fpath_list):
        if i % 100 == 0: print('Step ' + str(i))

        f = h5py.File(fpath, 'r')

        # tracks = f['mc_tracks']
        # tracks_neutr = tracks[tracks['bjorkeny'] != 0]
        # assert f['event_info'].shape == tracks_neutr.shape
        tracks_neutr = f['y']

        energies = tracks_neutr['energy']

        if cut_e_higher_than_3 is True:
            energies = energies[energies <= 3]

        if energy_conc_arr is None:
            energy_conc_arr = energies
        else:
            energy_conc_arr = np.concatenate([energy_conc_arr, energies], axis=0)

        f.close()

    print('Total number of events for ' + fpath_list_key_ic + ' : '
          + str(energy_conc_arr.shape[0]))
    print('Total number of files: ' + str(len(fpath_list)))

    return energy_conc_arr


def save_energies_for_ic(energies_for_ic):
    """

    Parameters
    ----------
    energies_for_ic

    Returns
    -------

    """

    np.savez('./energies_for_ic.npz',
             muon_cc_3_100=energies_for_ic['muon_cc_3_100'], muon_cc_1_5=energies_for_ic['muon_cc_1_5'],
             elec_cc_3_100=energies_for_ic['elec_cc_3_100'], elec_cc_1_5=energies_for_ic['elec_cc_1_5'],
             elec_nc_3_100=energies_for_ic['elec_nc_3_100'], elec_nc_1_5=energies_for_ic['elec_nc_1_5'])


def load_energies_for_ic():
    """

    Returns
    -------

    """

    data = np.load('./energies_for_ic.npz')

    energies_for_ic = dict()
    energies_for_ic['muon_cc_3_100'] = data['muon_cc_3_100']
    energies_for_ic['muon_cc_1_5'] = data['muon_cc_1_5']
    energies_for_ic['elec_cc_3_100'] = data['elec_cc_3_100']
    energies_for_ic['elec_cc_1_5'] = data['elec_cc_1_5']
    energies_for_ic['elec_nc_3_100'] = data['elec_nc_3_100']
    energies_for_ic['elec_nc_1_5'] = data['elec_nc_1_5']

    return energies_for_ic


def add_low_and_high_e_prods(energies_for_ic):
    """

    Parameters
    ----------
    energies_for_ic

    Returns
    -------

    """

    energies_for_ic['muon_cc'] = np.concatenate([energies_for_ic['muon_cc_3_100'], energies_for_ic['muon_cc_1_5']])
    energies_for_ic['elec_cc'] = np.concatenate([energies_for_ic['elec_cc_3_100'], energies_for_ic['elec_cc_1_5']])
    energies_for_ic['elec_nc'] = np.concatenate([energies_for_ic['elec_nc_3_100'], energies_for_ic['elec_nc_1_5']])
    energies_for_ic['elec_cc_and_nc'] = np.concatenate([energies_for_ic['elec_cc'], energies_for_ic['elec_nc']])


def plot_e_and_make_flat_func(energies_for_ic):
    """

    Parameters
    ----------
    energies_for_ic

    Returns
    -------

    """
    def make_plot_options_and_save(ax, pdfpages, ylabel):
        plt.xlabel('Energy [GeV]')
        plt.ylabel(ylabel)
        x_ticks_major = np.arange(0, 101, 10)
        ax.set_xticks(x_ticks_major)
        ax.grid(True)
        plt.tight_layout()
        pdfpages.savefig(fig)
        plt.cla()


    pdfpages = PdfPages('./e_hist_plots.pdf')
    fig, ax = plt.subplots()

    e_bins_1_to_2 = np.linspace(1, 2, 3)
    e_bins_2_to_25 = np.linspace(2.25, 25, 92)
    e_bins_25_to_60 = np.linspace(25.5, 60, 70)
    e_bins_60_to_100 = np.linspace(61, 100, 40)

    e_bins = np.concatenate([e_bins_1_to_2, e_bins_2_to_25, e_bins_25_to_60, e_bins_60_to_100], axis=0)

    # plot
    hist_muon_cc = plt.hist(energies_for_ic['muon_cc'], bins=e_bins)
    plt.title('Muon-CC 1-5 + 3-100 GeV for Run 1-2400')
    make_plot_options_and_save(ax, pdfpages, ylabel='Counts [#]')

    hist_shower = plt.hist(energies_for_ic['elec_cc_and_nc'], bins=e_bins)
    plt.title('Shower (elec-CC + elec-NC) 1-5 + 3-100 GeV for 2x Run 1-1200')
    make_plot_options_and_save(ax, pdfpages, ylabel='Counts [#]')

    hist_elec_cc = plt.hist(energies_for_ic['elec_cc'], bins=e_bins)
    plt.title('Elec-CC 1-5 + 3-100 GeV for Run 1-1200')
    make_plot_options_and_save(ax, pdfpages, ylabel='Counts [#]')

    hist_elec_nc = plt.hist(energies_for_ic['elec_nc'], bins=e_bins)
    plt.title('Elec-NC 1-5 + 3-100 GeV for Run 1-1200')
    make_plot_options_and_save(ax, pdfpages, ylabel='Counts [#]')

    # # We take 600 muon-CC files and 300 elec-cc and 300 elec_nc files for the split, reduce 1-3GeV bins by 1/2
    # hist_shower[0][0] = hist_shower[0][0] / 2 # 1-2GeV
    # hist_shower[0][1] = hist_shower[0][1] / 2 # 2-3GeV

    track_div_shower = np.divide(hist_muon_cc[0], hist_shower[0])
    # print(hist_muon_cc[0])
    # print(hist_shower[0])

    bins=hist_muon_cc[1] # doesnt matter which bins to use
    track_div_shower_mpl = np.append(track_div_shower, track_div_shower[-1])

    ax.step(bins, track_div_shower_mpl, linestyle='-', where='post')
    plt.title('Ratio tracks divided by showers')
    make_plot_options_and_save(ax, pdfpages, ylabel='Fraction')

    pdfpages.close()

    # save e_bins and corresponding fractions

    # make e_bins with mean center

    # e_bins = []
    # for i in range(bins.shape[0] - 1):
    #     e_mean = (bins[i] + bins[i+1]) / 2
    #     e_bins.append(e_mean)
    #
    # e_bins = np.array(e_bins)
    #
    #
    # arr_fract_for_e_bins = np.vstack((e_bins, track_div_shower))
    # print(arr_fract_for_e_bins)
    #
    # np.save('./arr_fract_for_e_bins.npy', arr_fract_for_e_bins)


def main():
    dirs_temp = {
            'muon_cc_3_100': '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/muon-CC',
            'muon_cc_1_5': '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/1-5GeV/muon-CC',
            'elec_cc_3_100': '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/elec-CC',
            'elec_cc_1_5': '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/1-5GeV/elec-CC',
            'elec_nc_3_100': '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/3-100GeV/elec-NC',
            'elec_nc_1_5': '/home/saturn/capn/mppi033h/Data/raw_data/ORCA_JTE_NEMOWATER/calibrated/with_jte_times/1-5GeV/elec-NC'
            }

    dirs = {
            'muon_cc_3_100': '/home/saturn/capn/mppi033h/Data/input_images/ORCA_2016_115l/tight_1_60b_ts_classifier/muon-CC/3-100GeV/xyzt',
            'muon_cc_1_5': '/home/saturn/capn/mppi033h/Data/input_images/ORCA_2016_115l/tight_1_60b_ts_classifier/muon-CC/1-5GeV/xyzt',
            'elec_cc_3_100': '/home/saturn/capn/mppi033h/Data/input_images/ORCA_2016_115l/tight_1_60b_ts_classifier/elec-CC/3-100GeV/xyzt',
            'elec_cc_1_5': '/home/saturn/capn/mppi033h/Data/input_images/ORCA_2016_115l/tight_1_60b_ts_classifier/elec-CC/1-5GeV/xyzt',
            'elec_nc_3_100': '/home/saturn/capn/mppi033h/Data/input_images/ORCA_2016_115l/tight_1_60b_ts_classifier/elec-NC/3-100GeV/xyzt',
            'elec_nc_1_5': '/home/saturn/capn/mppi033h/Data/input_images/ORCA_2016_115l/tight_1_60b_ts_classifier/elec-NC/1-5GeV/xyzt'
            }

    if os.path.isfile('./energies_for_ic.npz') is True:
        energies_for_ic = load_energies_for_ic()

    else:
        fpaths = dict()
        for dir_ic_key in dirs:
            fpaths[dir_ic_key] = get_h5_filepaths(dirs[dir_ic_key])

        energies_for_ic = dict()
        for fpath_list_key_ic in fpaths:
            print('Getting energies for ' + fpath_list_key_ic)
            cut_flag = False
            # cut_flag = True if fpath_list_key_ic in ['muon_cc_1_5', 'elec_cc_1_5', 'elec_nc_1_5'] else False
            fpath_list = fpaths[fpath_list_key_ic]
            energies_for_ic[fpath_list_key_ic] = get_energies_for_fpaths(fpath_list, fpath_list_key_ic, cut_e_higher_than_3=cut_flag)

        save_energies_for_ic(energies_for_ic)

    add_low_and_high_e_prods(energies_for_ic)
    plot_e_and_make_flat_func(energies_for_ic)


if __name__ == '__main__':
    main()


from unittest import TestCase
import numpy as np
import orcasong.modules as modules
from km3pipe.dataclasses import Table


__author__ = 'Stefan Reck'


class TestModules(TestCase):
    def test_mc_info_maker(self):
        """ Test the mcinfo maker on some dummy data. """
        def mc_info_extr(blob):
            hits = blob["Hits"]
            return {"dom_id_0": hits.dom_id[0],
                    "time_2": hits.time[2]}

        in_blob = {
            "Hits": Table({
                'dom_id': [2, 3, 3],
                'channel_id': [0, 1, 2],
                'time': [10.1, 11.2, 12.3]
            })
        }
        module = modules.McInfoMaker(
            mc_info_extr=mc_info_extr, store_as="test")
        out_blob = module.process(in_blob)

        self.assertSequenceEqual(list(out_blob.keys()), ["Hits", "test"])
        self.assertSequenceEqual(list(out_blob["test"].dtype.names),
                                 ('dom_id_0', 'time_2'))
        np.testing.assert_array_equal(out_blob["test"]["dom_id_0"],
                                      np.array([2, ]))
        np.testing.assert_array_equal(out_blob["test"]["time_2"],
                                      np.array([12.3, ]))

    def test_event_skipper(self):
        def event_skipper(blob):
            return blob == 42

        module = modules.EventSkipper(event_skipper=event_skipper)

        self.assertEqual(module.process(42), None)
        self.assertEqual(module.process(25), 25)


class TestTimePreproc(TestCase):
    def setUp(self):
        self.in_blob = {
            "Hits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        self.in_blob_mc = {
            "Hits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "McHits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

    def test_time_preproc_t0(self):
        module = modules.TimePreproc(
            add_t0=True, center_time=False)

        target = {
            "Hits": Table({
                'time': [1.1, 2.2, 3.3],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        out_blob = module.process(self.in_blob)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_equal(np.array(out_blob["Hits"]),
                                      np.array(target["Hits"]))

    def test_time_preproc_center(self):
        module = modules.TimePreproc(
            add_t0=False, center_time=True)

        target = {
            "Hits": Table({
                'time': [-1., 0., 1.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        out_blob = module.process(self.in_blob)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_equal(np.array(out_blob["Hits"]),
                                      np.array(target["Hits"]))

    def test_time_preproc_t0_and_center(self):
        module = modules.TimePreproc(
            add_t0=True, center_time=True)

        target = {
            "Hits": Table({
                'time': [-1.1, 0., 1.1],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        out_blob = module.process(self.in_blob)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))

    def test_time_preproc_mchits_t0_and_center(self):
        module = modules.TimePreproc(
            add_t0=True, center_time=True)

        target = {
            "Hits": Table({
                'time': [-1.1, 0., 1.1],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "McHits": Table({
                'time': [-1.1, 0., 1.1],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
        }
        out_blob = module.process(self.in_blob_mc)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["McHits"].view("<f8")),
            np.array(target["McHits"].view("<f8")))


class TestImageMaker(TestCase):
    def test_2d_xt_binning(self):
        # (3 x 2) x-t binning
        bin_edges_list = [
            ["x", [3.5, 4.5, 5.5, 6.5]],
            ["time", [0.5, 2, 3.5]]
        ]

        module = modules.ImageMaker(
            bin_edges_list=bin_edges_list, store_as="histogram")
        in_blob = {
            "Hits": Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        target = {
            "Hits": Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "histogram": np.array([[
                [1, 0],
                [0, 1],
                [0, 1],
            ]])
        }

        out_blob = module.process(in_blob)
        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["histogram"]),
            np.array(target["histogram"]))

    def test_unknown_field(self):
        # (3 x 2) x-t binning
        bin_edges_list = [
            ["aggg", [3.5, 4.5, 5.5, 6.5]],
            ["time", [0.5, 2, 3.5]]
        ]

        module = modules.ImageMaker(
            bin_edges_list=bin_edges_list, store_as="histogram")
        in_blob = {
            "Hits": Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        with self.assertRaises(ValueError):
            module.process(in_blob)

    def test_1d_binning(self):
        # (1, ) t binning
        bin_edges_list = [
            ["time", [2.5, 3.5]]
        ]

        module = modules.ImageMaker(
            bin_edges_list=bin_edges_list, store_as="histogram")
        in_blob = {
            "Hits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        target = {
            "Hits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "histogram": np.array([
                [1, ],
            ])
        }

        out_blob = module.process(in_blob)
        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["histogram"]),
            np.array(target["histogram"]))

    def test_1d_binning_no_hits(self):
        # (1, ) t binning
        bin_edges_list = [
            ["time", [3.5, 4.5]]
        ]

        module = modules.ImageMaker(
            bin_edges_list=bin_edges_list, store_as="histogram")
        in_blob = {
            "Hits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        target = {
            "Hits": Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "histogram": np.array([
                [0, ],
            ])
        }

        out_blob = module.process(in_blob)
        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["histogram"]),
            np.array(target["histogram"]))


class TestBinningStatsMaker(TestCase):
    def test_it(self):
        # (3 x 2) x-t binning
        bin_edges_list = [
            ["x", [3.5, 4.5, 5.5, 6.5]],
            ["time", [0.5, 2, 3.5]],
            ["z", [1, 4]]
        ]

        in_blob = {
            "Hits": Table({
                "x": [4, 5, 6, 6],
                'time': [1., 2., 3., 50],
                "z": [0, 3, 4, 5],

                "t0": [0.1, 0.2, 0.3, 0.4],
                "triggered": [0, 1, 1, 1],
            })
        }

        target = {
            'x': {
                'hist': np.array([0., 0., 0., 1., 0., 1.]),
                'hist_bin_edges': np.array([3.5, 4., 4.5, 5., 5.5, 6., 6.5]),
                'bin_edges': [3.5, 4.5, 5.5, 6.5],
                'cut_off': np.array([0., 0.])
            },
            'time': {
                'hist': np.array([0., 2.]),
                'hist_bin_edges': [0.5, 2, 3.5],
                'bin_edges': [0.5, 2, 3.5],
                'cut_off': np.array([0., 1.])
            },
            'z': {
                'hist': np.array([0., 2.]),
                'hist_bin_edges': np.array([1., 2.5, 4.]),
                'bin_edges': [1, 4],
                'cut_off': np.array([1., 1.])
            }
        }

        module = modules.BinningStatsMaker(
            bin_edges_list=bin_edges_list, res_increase=2)
        module.process(in_blob)
        output = module.finish()
        check_dicts_n_ray(output, target)


def check_dicts_n_ray(a, b):
    """ Check if dicts with dicts with ndarrays are equal. """
    if set(a.keys()) != set(b.keys()):
        raise KeyError("{} != {}".format(a.keys(), b.keys()))
    for key in a.keys():
        if set(a[key].keys()) != set(b[key].keys()):
            raise KeyError("{} != {}".format(a[key].keys(), b[key].keys()))
        for skey in a[key].keys():
            np.testing.assert_array_almost_equal(a[key][skey], b[key][skey])



"""
Run with a parser to plot the binning statistics.
Functions for plotting the bin stats made by the BinningStatsMaker module.
"""

import os
import warnings
import argparse
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import h5py
import numpy as np


__author__ = 'Stefan Reck'


def plot_hists(hists, save_to, plot_bin_edges=True):
    """
    Plot histograms made by the BinningStatsMaker to the given pdf path.

    Parameters
    ----------
    hists : dict or List
        Dicts with the info about the binning, generated by the BinningStatsMaker.
        Can also be a list of dicts (for multiple files), which will be
        combined to one plot.
    save_to : str
        Where to save the plot to.
    plot_bin_edges : bool
        If true, will plot the bin edges as horizontal lines. Is never used
        for the time binning (field name "time").

    """
    if isinstance(hists, list):
        hists = combine_hists(hists)

    with PdfPages(save_to) as pdf_file:
        for bin_name, hists_data in hists.items():
            hist_bin_edges = hists_data["hist_bin_edges"]
            bin_edges = hists_data["bin_edges"]
            hist = hists_data["hist"]
            cut_off = hists_data["cut_off"]
            total_hits = np.sum(hist) + np.sum(cut_off)

            bin_widths = np.diff(hist_bin_edges)
            hist_prob = hist / bin_widths / np.sum(hist)

            fig, ax = plt.subplots()
            plt.bar(hist_bin_edges[:-1],
                    hist_prob,
                    align="edge",
                    width=bin_widths
                    )

            if plot_bin_edges and bin_name != "time":
                for bin_edge in bin_edges:
                    plt.axvline(x=bin_edge, color='grey', linestyle='-',
                                linewidth=1, alpha=0.9)

            # place a text box in upper left in axes coords
            out_neg_rel = cut_off[0] / total_hits
            out_pos_rel = cut_off[1] / total_hits
            textstr = "Hits cut off:\nLeft: {:.1%}\n" \
                      "Right: {:.1%}".format(out_neg_rel, out_pos_rel)
            props = dict(boxstyle='round', facecolor='white', alpha=0.9)
            ax.text(0.95, 0.95, textstr, transform=ax.transAxes,
                    verticalalignment='top', bbox=props,
                    horizontalalignment="right")

            # the auto ticks are nice even numbers
            ylims = [0, ax.get_yticks().max()]
            ax.set_ylim(ylims)

            plt.xlabel(bin_name)
            plt.ylabel("Density of hits")

            pdf_file.savefig(fig)

    print("Saved binning plot to " + save_to)


def combine_hists(hists_list):
    """
    Combine the hists for multiple files into a single big one.

    Parameters
    ----------
    hists_list : List
        A list of dicts (hists from the BinningStatsMaker for different files).

    Returns
    -------
    combined_hists : Dict
        The combined hist.

    """
    # initialize combined dict according to first dict in list
    combined_hists = {}
    for bin_name, hists_data in hists_list[0].items():
        hist_bin_edges = hists_data["hist_bin_edges"]
        bin_edges = hists_data["bin_edges"]
        bin_name = bin_name

        combined_hists[bin_name] = {
            "hist": np.zeros(len(hist_bin_edges) - 1),
            "hist_bin_edges": hist_bin_edges,
            "bin_edges": bin_edges,
            # below smallest edge, above largest edge:
            "cut_off": np.zeros(2),
        }

    # add them together
    for hists in hists_list:
        for bin_name, hists_data in hists.items():
            # name and bin edges must be the same for all hists
            if bin_name not in combined_hists:
                raise NameError(
                    "Hists dont all have the same binning field name:"
                    "{}".format(bin_name))

            bin_edges = hists_data["bin_edges"]
            comb_edges = combined_hists[bin_name]["bin_edges"]
            if len(bin_edges) != len(comb_edges):
                raise ValueError("Hists have different hist bin edges: {} vs {}"
                                 .format(bin_edges, comb_edges))

            hist_bin_edges = hists_data["hist_bin_edges"]
            comb_hist_edges = combined_hists[bin_name]["hist_bin_edges"]
            if len(hist_bin_edges) != len(comb_hist_edges):
                raise ValueError("Hists have different bin edges: {} vs {}"
                                 .format(hist_bin_edges, comb_hist_edges))

            combined_hists[bin_name]["hist"] += hists_data["hist"]
            combined_hists[bin_name]["cut_off"] += hists_data["cut_off"]

    return combined_hists


def add_hists_to_h5file(hists, file):
    """
    Add the binning stats as groups to a h5 file.

    Parameters
    ----------
    hists : dict
        The histst from the BinningStatsMaker module.
    file : str
        Path to the h5 file.

    """
    base_name = "bin_stats/"
    with h5py.File(file, "a") as f:
        for bin_name, hists_data in hists.items():
            for key, val in hists_data.items():
                h5_folder = base_name + bin_name + "/" + key
                f.create_dataset(h5_folder, data=val)


def plot_hist_of_files(save_as, files=None):
    """
    Plot the binning stats of all h5 files in the current directory.

    Parameters
    ----------
    save_as : str
        Where to save the plot to.
    files : List, optional
        Path of files to use instead.

    """
    if files is None:
        files = get_all_h5_files()

    hists_list = []
    opened_files = []
    print("Plotting stats of {} file(s)".format(len(files)))

    try:
        print("Opening files...")
        for file in files:
            f = h5py.File(file, "r")
            if "bin_stats/" not in f:
                warnings.warn("ERROR: File {} does not have bin_stats dataset. "
                              "Skipping ...".format(file))
                f.close()
                continue
            hists_list.append(f["bin_stats/"])
            opened_files.append(f)

        print("Plotting...")
        plot_hists(hists_list, save_as)

    finally:
        print("Closing files...")
        for file in opened_files:
            file.close()


def get_all_h5_files():
    """ Get a list of all h5 files in the cwd. """
    files = []
    for file in os.listdir(os.getcwd()):
        if file.endswith(".h5"):
            files.append(file)
    return files


def main():
    parser = argparse.ArgumentParser(
        description='Generate a plot with statistics of the binning. '
                    'Can only be used on files generated with the FileBinner when '
                    'add_bin_stats was set to true (default). ')

    parser.add_argument('save_as', type=str, nargs="?",
                        default="bin_stats_plot.pdf",
                        help='Filename of the plot. Default: '
                             'bin_stats_plot.pdf.')

    parser.add_argument('file', type=str, nargs='*', default=None,
                        help='File(s) to plot. Default: Plot for all h5 '
                             'files in current dir.')

    args = parser.parse_args()
    plot_hist_of_files(args.save_as, args.file)


if __name__ == "__main__":
    main()
